@@ -3,8 +3,8 @@ import numpy as np
 import os
 import json
 
-from assembled.compatibility.faked_classifier import probability_calibration_for_faked_models, _initialize_fake_models
 from assembled.utils.logger import get_logger
+
 from typing import List, Tuple, Optional, Callable, Union
 from sklearn.model_selection import train_test_split
 
@@ -1034,7 +1034,7 @@ class MetaTask:
 
             self.remove_predictors(list(predictors_to_remove))
 
-    # -- Run/Benchmark Functions
+    # -- Data Access Functions
     def get_indices_for_fold(self, fold_idx, return_indices=False):
         train_indices = self.folds != fold_idx
         test_indices = self.folds == fold_idx
@@ -1122,378 +1122,61 @@ class MetaTask:
                meta_dataset.loc[:, validation_predictions_columns], \
                meta_dataset.loc[:, validation_confidences_columns]
 
-    @staticmethod
-    def _save_fold_results(y_true, y_pred, fold_idx, out_path, technique_name, classification=True):
-        """Code to Save Fold results in a file such that we can add any other folds results in the future"""
-        if out_path is None:
-            return
-
-        # Path test
-        path_exists = os.path.exists(out_path)
-
-        # Get data that is to be saved
-        index_metatask = y_true.index
-        fold_indicator = [fold_idx for _ in range(len(y_true))]
-        res_df = pd.DataFrame(np.array([y_true, index_metatask, fold_indicator, y_pred]).T,
-                              columns=["ground_truth", "Index-Metatask", "Fold", technique_name])
-
-        # Keep type correct
-        if classification:
-            # Note: while we tell read_csv that technique name should be read as string, it does not matter if
-            #   technique name is not in the file to be read. Hence it works for loading it initially
-            input_dtype = "string"
-            res_df = res_df.astype(input_dtype).astype({"Index-Metatask": int, "Fold": int})
-        else:
-            input_dtype = None
-
-        # Save folds base
-        if fold_idx == 0:
-            if path_exists:
-                # Load old data
-                tmp_df = pd.read_csv(out_path, dtype=input_dtype).astype(
-                    {"Index-Metatask": int, "Fold": int}).set_index(["ground_truth", "Index-Metatask", "Fold"])
-                # Pre-fill technique values
-                tmp_df[technique_name] = pd.Series(np.nan, index=np.arange(len(tmp_df)))
-                # Insert new values into data
-                res_df = res_df.set_index(["ground_truth", "Index-Metatask", "Fold"])
-
-                # Try-catch to detect (potential) random seed error
-                try:
-                    tmp_df.loc[res_df.index, technique_name] = res_df
-                except KeyError:
-                    raise KeyError("Wrong fold/index/ground-truth combinations found in results file." +
-                                   " Did you create the existing results file with a different random seed for" +
-                                   " the meta train test split? " +
-                                   " To fix this delete the existing files or adjust the random seed.")
-
-                tmp_df.reset_index().sort_values(by=["Index-Metatask"]).to_csv(out_path, index=False)
-            else:
-                res_df.to_csv(out_path, index=False)
-
-        else:
-            # Check error
-            if not path_exists:
-                raise RuntimeError("Somehow a later fold is trying to be saved first. " +
-                                   "Did the output file got deleted between folds?")
-
-            # Read so-far data
-            tmp_df = pd.read_csv(out_path, dtype=input_dtype).astype(
-                {"Index-Metatask": int, "Fold": int}).set_index(["ground_truth", "Index-Metatask", "Fold"])
-
-            if len(list(tmp_df)) == 1:
-                # Initial case, first time file building
-                res_df.to_csv(out_path, mode='a', header=False, index=False)
-            else:
-                # Default case later, same as above wihtout init
-                res_df = res_df.set_index(["ground_truth", "Index-Metatask", "Fold"])
-                tmp_df.loc[res_df.index, technique_name] = res_df
-                tmp_df.reset_index().sort_values(by=["Index-Metatask"]).to_csv(out_path, index=False)
-
-    def run_ensemble_on_all_folds(self, technique, technique_args: dict, technique_name,
-                                  use_validation_data_to_train_ensemble_techniques: bool = False,
-                                  meta_train_test_split_fraction: float = 0.5, meta_train_test_split_random_state=0,
-                                  pre_fit_base_models: bool = False, base_models_with_names: bool = False,
-                                  label_encoder=False, preprocessor=None, output_file_path=None, oracle=False,
-                                  probability_calibration="no", return_scores: Optional[Callable] = None):
-        """Run an ensemble technique on all folds and return the results
-
-        The current implementation builds fake base models by default such that we can evaluate methods this way.
-
-        !Warning!: we will overwrite older data if the technique name is already existing in the results file!
-        We will also not delete an existing file. If a file exist, we only add to/overwrite the file.
+    def yield_evaluation_data(self, folds_to_run: Optional[List[int]] = None) -> Tuple[int, pd.DataFrame, pd.DataFrame,
+                                                                                       pd.Series, pd.Series,
+                                                                                       pd.DataFrame, pd.DataFrame,
+                                                                                       pd.DataFrame, pd.DataFrame]:
+        """Yield the dataset and base model data for the specified folds.
 
         Parameters
         ----------
-        technique: sklearn estimator like object
-            The Ensemble Technique (as a sklearn like fit/predict style) model.
-        technique_args: dict
-            The arguments that shall be supplied to the ensemble
-        technique_name: str
-            Name of the technique used to identify the technique later on in the saved results.
-        use_validation_data_to_train_ensemble_techniques: bool, default=False
-            Whether to use validation data to train the ensemble techniques (through the faked base models) and use the
-            fold's prediction data to evaluate the ensemble technique.
-            If True, the metataks requires validation data. If False, test predictions of a fold are split into
-            meta_train and meta_test subsets where the meta_train subset is used to train the ensemble techniques and
-            meta_test to evaluate the techniques.
-        meta_train_test_split_fraction: float, default=0.5
-            The fraction for the meta train/test split. Only used if
-        meta_train_test_split_random_state: int, default=0
-            The randomness for the meta train/test split.
-        pre_fit_base_models: bool, default=False
-            Whether or not the base models need to be fitted to be passed to the ensemble technique.
-        base_models_with_names: bool, default=False
-            Whether or not the base models' list should contain the model and its name.
-        label_encoder: bool, default=False
-            Whether the ensemble technique expects that a label encoder is applied to the fake models. Often required
-            for sklearn ensemble techniques.
-        preprocessor: sklearn-like transformer, default=None
-            Function used to preprocess the data for later. called fit_transform on X_train and transform on X_test.
-        output_file_path: str, default=None
-            File path where the results of the folds shall be stored. If none, we do not store anything.
-            We assume the file is in the correct format if it exists and will create it if it does not exit.
-            Here, no option to purge/delete existing files is given. This is must be done in an outer scope.
-        oracle: bool, default=False
-            Whether the ensemble technique is an oracle. If true, we pass and call the method differently.
-        probability_calibration: {"sigmoid", "isotonic", "auto", "no"}, default="no"
-            What type of probability calibration (see https://scikit-learn.org/stable/modules/calibration.html)
-            shall be applied to the base models:
+        folds_to_run: List of int, default=None
+            If None, yield data for all folds.
+            If not None, the function will only return the fold data for the fold indices
+            specified in the list.
 
-                - "sigmoid": Use CalibratedClassifierCV with method="sigmoid"
-                - "isotonic": Use CalibratedClassifierCV with method="isotonic"
-                - "auto": Determine which method to use for CalibratedClassifierCV depending on the number of instances.
-                - "no": Do not use probability calibration.
-
-            If pre_fit_base_models is False, CalibratedClassifierCV is employed with ensemble="False" to simulate
-            cross_val_predictions by our Faked Base Models.
-            If pre_fit_base_models is True, CalibratedClassifierCV is employed with cv="prefit" beforehand such that
-            we "replace" the base models with calibrated base models.
-        return_scores: Callable, default=None
-            If the evaluation shall return the scores for each fold. If not None, a metric function is expected.
+        Yields for each fold
+        ----------
+        idx: int
+            Fold index
+        X_train: DataFrame
+            Feature Data used to train base models
+        X_test: DataFrame
+            Feature Data used to test base models
+        y_train: Series
+            Label Data used to train base models
+        y_test: Series
+            Label Data used to test base models
+        val_base_predictions: DataFrame
+            Predictions of each base model on the fold's validation data (if exists)
+        test_base_predictions: DataFrame
+            Predictions of each base model on the fold's test data
+        val_base_confidences: DataFrame
+            Confidences of each base model on the fold's validation data (if exists)
+        test_base_confidences: DataFrame
+            Confidences of each base model on the fold's test data
         """
-        # TODO -- Parameter Preprocessing / Checking
-        #   Add safety check for file path here or something
-        #   Check if probability_calibration has correct string names
-        #   Check metric / scorer object
-        if use_validation_data_to_train_ensemble_techniques and (not self.use_validation_data):
-            raise ValueError("Metatask has no validation data but use_validation_to_train_ensemble_techniques is True.")
 
-        # -- Re-names
-        val_data = use_validation_data_to_train_ensemble_techniques
-        m_frac = meta_train_test_split_fraction
-        m_rand = meta_train_test_split_random_state
-
-        # -- Iterate over Folds
-        fold_scores = []
         for idx, train_metadata, test_metadata in self.fold_split(return_fold_index=True):
+            # -- Check if Fold is wanted, if not skip
+            if (folds_to_run is not None) and (idx not in folds_to_run):
+                continue
 
-            # -- Get Data from Metatask
             X_train, y_train, _, _, val_base_predictions, val_base_confidences = self.split_meta_dataset(train_metadata,
                                                                                                          fold_idx=idx)
             X_test, y_test, test_base_predictions, test_base_confidences, _, _ = self.split_meta_dataset(test_metadata,
                                                                                                          fold_idx=idx)
 
-            # -- Employ Preprocessing and input validation
-            train_indices = X_train.index.to_numpy()
-            X_train, X_test, y_train, y_test = self._check_fold_data_for_ensemble(X_train, X_test, y_train, y_test,
-                                                                                  preprocessor)
-            # -- Get Data for Fake Base Model and Evaluation
-            # - Data on which the base models have been trained and had to predict for the fold
-            test_base_model_train_X = X_train
-            test_base_model_train_y = y_train
-
-            # - Validation Data
-            if not val_data:
-                ensemble_train_X, ensemble_test_X, ensemble_train_y, ensemble_test_y, fake_base_model_known_X, \
-                fake_base_model_known_predictions, fake_base_model_known_confidences, val_base_model_train_X, \
-                val_base_model_train_y = self._get_data_for_ensemble_without_validation(X_train, X_test, y_train,
-                                                                                        y_test,
-                                                                                        test_base_predictions,
-                                                                                        test_base_confidences,
-                                                                                        m_frac, m_rand)
-            else:
-                # Get iloc indices of validation data (Default value is for backwards compatibility)
-                fold_validation_indices = self.validation_indices.get(idx, train_indices)
-                # iloc_indices corresponds to np.arange(len(X_test)) for cross-validation
-                iloc_indices = np.intersect1d(fold_validation_indices, train_indices, return_indices=True)[2]
-
-                # Get Data
-                ensemble_train_X, ensemble_test_X, ensemble_train_y, ensemble_test_y, fake_base_model_known_X, \
-                fake_base_model_known_predictions, fake_base_model_known_confidences, val_base_model_train_X, \
-                val_base_model_train_y = self._get_data_for_ensemble_with_validation(X_train[iloc_indices], X_test,
-                                                                                     y_train[iloc_indices], y_test, idx,
-                                                                                     val_base_predictions.iloc[
-                                                                                         iloc_indices],
-                                                                                     val_base_confidences.iloc[
-                                                                                         iloc_indices],
-                                                                                     test_base_predictions,
-                                                                                     test_base_confidences)
-
-            # -- Build Fake Base Models
-            base_models = self._build_fake_base_models(test_base_model_train_X, test_base_model_train_y,
-                                                       val_base_model_train_X, val_base_model_train_y,
-                                                       ensemble_train_X, ensemble_train_y,
-                                                       fake_base_model_known_X,
-                                                       fake_base_model_known_predictions,
-                                                       fake_base_model_known_confidences,
-                                                       pre_fit_base_models, base_models_with_names,
-                                                       label_encoder, probability_calibration)
-
-            # -- Run Ensemble for Fold
-            y_pred_ensemble_model = self._run_ensemble_on_data(base_models, technique, technique_args,
-                                                               ensemble_train_X, ensemble_test_X, ensemble_train_y,
-                                                               ensemble_test_y, oracle)
-
-            # -- Post Process Results
-            self._save_fold_results(ensemble_test_y, y_pred_ensemble_model, idx, output_file_path, technique_name)
-
-            # -- Save scores for return
-            if return_scores is not None:
-                fold_scores.append(return_scores(ensemble_test_y, y_pred_ensemble_model))
-
-        # -- Return Score
-        if return_scores is not None:
-            return fold_scores
-
-    def _build_fake_base_models(self, test_base_model_train_X, test_base_model_train_y,
-                                val_base_model_train_X, val_base_model_train_y,
-                                ensemble_train_X, ensemble_train_y, fake_base_model_known_X,
-                                fake_base_model_known_predictions, fake_base_model_known_confidences,
-                                pre_fit_base_models, base_models_with_names, label_encoder, probability_calibration):
-        """Build fake base models from data.
-
-        New Parameters (rest can be found in signature of run_ensemble_on_all_folds)
-        ----------
-        test_base_model_train_X: array-like
-            The instances on which the base models are trained for the (outer) evaluation
-        test_base_model_train_y: array-like
-            The labels on which the base models are trained for the (outer) evaluation
-        val_base_model_train_X:
-            The instances on which the base models are trained for the validation data.
-        val_base_model_train_y: array-like
-            The labels on which the base models are trained for the validation data.
-        ensemble_train_X: array-like
-            The instances on which the ensemble technique is to be trained.
-        ensemble_train_y: array-like
-            The labels on which the ensemble technique is to be trained.
-        fake_base_model_known_X: array-like
-            The instances from the test or training data for which the base model must know predictions.
-        fake_base_model_known_predictions: pd.DataFrame
-            Predictions for the instances of fake_base_model_known_X
-        fake_base_model_known_confidences: pd.DataFrame
-            Confidences for the instances of fake_base_model_known_X
-        """
-
-        # -- Build ensemble technique
-        base_models = _initialize_fake_models(test_base_model_train_X, test_base_model_train_y,
-                                              val_base_model_train_X, val_base_model_train_y, fake_base_model_known_X,
-                                              fake_base_model_known_predictions, fake_base_model_known_confidences,
-                                              pre_fit_base_models, base_models_with_names, label_encoder,
-                                              self.to_confidence_name)
-
-        # -- Probability Calibration
-        base_models = probability_calibration_for_faked_models(base_models, ensemble_train_X, ensemble_train_y,
-                                                               probability_calibration, pre_fit_base_models)
-
-        return base_models
-
-    @staticmethod
-    def _run_ensemble_on_data(base_models, technique, technique_args, ensemble_train_X, ensemble_test_X,
-                              ensemble_train_y, ensemble_test_y, oracle):
-        """Run an ensemble technique for the given data and return its predictions
-
-        New Parameters (rest can be found in signature of run_ensemble_on_all_folds or _build_fake_base_models)
-        ----------
-        base_models: List of fake base models
-            A list of base models that shall be used by the ensemble technique.
-        ensemble_test_X: array-like
-            The instances on which the ensemble technique shall predict on.
-        ensemble_test_y: array-like
-            The labels which the ensemble technique shall predict (only used for oracle-like techniques).
-        """
-
-        # -- Build ensemble model
-        ensemble_model = technique(base_models, **technique_args)
-
-        # -- Fit and Predict
-        ensemble_model.fit(ensemble_train_X, ensemble_train_y)
-
-        if oracle:
-            # Special case for virtual oracle-like predictors
-            y_pred_ensemble_model = ensemble_model.oracle_predict(ensemble_test_X, ensemble_test_y)
-        else:
-            y_pred_ensemble_model = ensemble_model.predict(ensemble_test_X)
-
-        return y_pred_ensemble_model
-
-    @staticmethod
-    def _check_fold_data_for_ensemble(X_train, X_test, y_train, y_test, preprocessor):
-        if preprocessor is not None:
-            X_train = preprocessor.fit_transform(X_train)
-            X_test = preprocessor.transform(X_test)
-        else:
-            # make sure X_train and X_test are arrays from here onwards
-            X_train = X_train.to_numpy()
-            X_test = X_test.to_numpy()
-
-        # Make sure y_train and y_test are arrays
-        y_train = y_train.to_numpy()
-        y_test = y_test.to_numpy()
-
-        return X_train, X_test, y_train, y_test
-
-    # -- Data for Ensemble Getters
-    @staticmethod
-    def _get_data_for_ensemble_without_validation(X_train, X_test, y_train, y_test, test_base_predictions,
-                                                  test_base_confidences, meta_train_test_split_fraction,
-                                                  meta_train_test_split_random_state):
-        """Split for ensemble technique evaluation only on fold predictions to get trainings data for the ensemble
-        """
-
-        # -- In this case, the base models have been fitted on the same data for the validation and test data
-        #   (because we take a subset of the test data as validation data)
-        val_base_model_train_X = X_train
-        val_base_model_train_y = y_train
-
-        # -- Data on the predictions of the original base model
-        fake_base_model_known_X = X_test
-        fake_base_model_known_predictions = test_base_predictions
-        fake_base_model_known_confidences = test_base_confidences
-
-        # Split of the original test data corresponding to the parts of the predictions used for train and test
-        ensemble_train_X, ensemble_test_X, ensemble_train_y, ensemble_test_y \
-            = train_test_split(X_test, y_test, test_size=meta_train_test_split_fraction,
-                               random_state=meta_train_test_split_random_state, stratify=y_test)
-
-        return ensemble_train_X, ensemble_test_X, ensemble_train_y, ensemble_test_y, fake_base_model_known_X, \
-               fake_base_model_known_predictions, fake_base_model_known_confidences, val_base_model_train_X, \
-               val_base_model_train_y
-
-    def _get_data_for_ensemble_with_validation(self, X_train, X_test, y_train, y_test, fold_idx,
-                                               val_base_predictions, val_base_confidences,
-                                               test_base_predictions, test_base_confidences):
-        """Build data for validation based evaluation
-
-        Fake Model must be able to predict for all instances of the validation data and the fold's predictions.
-        Hence, it must get all the data needed for that.
-        """
-
-        # -- Select the relevant subset of training instances that are part of the validation data
-        val_base_model_train_X = X_train
-        val_base_model_train_y = y_train
-
-        # -- Get known/oracle data for fake base model
-        # Therefore, combine validation and test instances
-        fake_base_model_known_X = np.vstack((val_base_model_train_X, X_test))
-
-        # Also need to combine predictions / confidences (keep as DF for columns)
-        # Have to rename columns to use DFs with validation data, here remove fold postfix to achieve this
-        p_rename = {self.to_validation_predictor_name(col_name, fold_idx): col_name for col_name
-                    in list(test_base_predictions)}
-        c_rename = {self.to_validation_predictor_name(col_name, fold_idx): col_name for col_name
-                    in list(test_base_confidences)}
-        fake_base_model_known_predictions = pd.concat([val_base_predictions.rename(columns=p_rename),
-                                                       test_base_predictions], axis=0)
-        fake_base_model_known_confidences = pd.concat([val_base_confidences.rename(columns=c_rename),
-                                                       test_base_confidences], axis=0)
-
-        # Finally set which subsets are used for what
-        # (These correspond to the subsets passed to the base model to get known predictions/confidences)
-        ensemble_train_X = val_base_model_train_X
-        ensemble_train_y = val_base_model_train_y
-        ensemble_test_X = X_test
-        ensemble_test_y = y_test
-
-        # Return
-        return ensemble_train_X, ensemble_test_X, ensemble_train_y, ensemble_test_y, fake_base_model_known_X, \
-               fake_base_model_known_predictions, fake_base_model_known_confidences, val_base_model_train_X, \
-               val_base_model_train_y
+            yield idx, X_train, X_test, y_train, y_test, val_base_predictions, test_base_predictions, \
+                  val_base_confidences, test_base_confidences
 
     # ---- Experimental Functions
     def _exp_yield_evaluation_data_across_folds(self, meta_train_test_split_fraction,
                                                 meta_train_test_split_random_state,
                                                 pre_fit_base_models, base_models_with_names, label_encoder,
                                                 preprocessor, include_test_data=False):
+        from assembled.compatibility.faked_classifier import _initialize_fake_models
+
         # FIXME, extend to validation data usage?
         for idx, train_metadata, test_metadata in self.fold_split(return_fold_index=True):
             X_train, y_train, _, _, _, _ = self.split_meta_dataset(train_metadata)
